# coding: utf-8

"""Partial raft consensus algorithm.

Includes the leader/follower selection aspect.

Currently, the log part is omitted, as we are looking only at
"who is the leader?" for database status messages.  However, we might
want to do the log anyway and have that be the framework for the
DB status messages themselves, which would eliminate the possibility
of any disagreement in status message, though this would increase latency
in getting the fastest DB status as we need to wait for agreement that
messages can first be committed.

This algo rides on top of the "rpc / peer" model given in rpc.py / p2p.py,
and is otherwise agnostic of the transport system in use.


"""
import logging
import random
import time

from . import rpc
from . import p2p
from . import state as _state
from . import util
from .util import event

log = logging.getLogger(__name__)

BATCH_SIZE = 1


class RaftListener(object):
    def status_changed(self, rafthandler, status):
        pass


class RaftHandler(object):
    name = 'raft'

    # slower election timer when things are connected
    # in a stable state, to greatly reduce message overhead and CPU
    # (verified with top this thing burns a lot when it's running
    # faster)
    smooth_sailing_range = (10, 15)

    # fast election timers when things are changing, so we can
    # reelect a leader quickly.
    freakout_range = (.150, .300)

    election_timeout_range = smooth_sailing_range
    rpc_timeout = .5

    state = _state.StateMachine()
    START = state.START
    LEADER = state.new("LEADER")
    FOLLOWER = state.new("FOLLOWER")
    CANDIDATE = state.new("CANDIDATE")

    def __init__(self, service):
        self.service = service
        self.p2p = p2p.PeerRunner(self.service)
        self.p2p.speak_rpc(rpc_reg)

        event.event_listen(self.service, "before_listen", self.before_listen)
        self.service.memos['raft'] = self
        self.state.init_instance(self)
        self.state.add_listener(self, self._state_changed)
        self.async_suite = service.async_suite
        self.candidate_id = service.nodename
        self.current_term = 0
        self.voted_for = None
        self._sped_up_at = float("inf")
        self.election_timer = None
        self.current_leader = None
        self._monitor_lock = self.async_suite.lock()

        event.event_listen(self.p2p, "new_peer", self._new_peer)
        event.event_listen(self.p2p, "peer_connected", self._peer_connected)
        event.event_listen(
            self.p2p, "peer_disconnected", self._peer_disconnected)

    def _state_changed(self, target, old_state, new_state):
        self.dispatch.state_changed(self, old_state, new_state)

    @property
    def is_leader(self):
        return self.current_leader == self.candidate_id

    def _new_peer(self, peer):
        """event hook from p2p runner"""
        peer.memos['voted_for_us'] = False
        peer.memos['ping_due'] = 0
        peer.memos['vote_due'] = 0

    def _peer_connected(self, peer):
        self._speed_up()
        self._run_monitor()

    def _peer_disconnected(self, peer):
        # TODO: check if we should see if only the leader here
        self._speed_up()
        self._run_monitor()

    def _update_memo(self, key, value):
        for peer in self.p2p.peers:
            peer.memos[key] = value

    def before_listen(self, server):
        self._raft_started = time.time()
        self.FOLLOWER.go(self)
        self._monitor_greenlet = self.async_suite.spawn(self._monitor)

    def _monitor(self):
        while True:
            self._run_monitor()
            self.async_suite.sleep(float(self.election_timeout_range[0]) / 4)

    def _run_monitor(self):
        if not self._monitor_lock.acquire(False):
            return
        try:
            now = time.time()
            if self.FOLLOWER.current(self):
                if self.election_timer(now):
                    self.CANDIDATE.go(self)
                else:
                    # slow down election timer after some period
                    self._check_slowdown(now, 20)
            elif self.LEADER.current(self):
                # slow down election timer after some period that
                # is considerably longer than what the followers do
                self._check_slowdown(now, 60)
                for peer in self.p2p.peers:
                    if not peer.connected:
                        continue
                    if peer.memos['ping_due'] < now:
                        self.request_append_entries(peer, now)
            elif self.CANDIDATE.current(self):
                if self.election_timer(now):
                    self.CANDIDATE.go(self)
                else:
                    any_ = False
                    for peer in self.p2p.peers:
                        if not peer.connected:
                            continue
                        any_ = True
                        if peer.memos['vote_due'] < now:
                            self.request_vote_request(peer, now)
                    if not any_:
                        self._check_votes()
        finally:
            self._monitor_lock.release()

    def _reset_election_timer(self, fire=False):
        timeout = random.uniform(*self.election_timeout_range)
        now = time.time()
        if fire:
            # make the first hit very soon
            now -= self.election_timeout_range[0]
        self.election_timer = util.periodic_timer(timeout, now)

    def _speed_up(self):
        reset_timer = self._sped_up_at == float("inf")
        self._sped_up_at = time.time()
        if reset_timer:
            self.election_timeout_range = self.freakout_range
            self._reset_election_timer(fire=True)
            log.info(
                "Peer changes detected, speeding up "
                "election timeouts to %.4f" % self.election_timeout_range[0])

    def _check_slowdown(self, now, gap_factor):
        gap = self.freakout_range[-1] * gap_factor
        if now - self._sped_up_at > gap:
            self.election_timeout_range = self.smooth_sailing_range
            self._reset_election_timer()
            log.info(
                "Entering stable state, slowing down "
                "election timeouts to %.4f" % self.election_timeout_range[0])
            self._sped_up_at = float("inf")

    @state.transition(START, FOLLOWER)
    def _start_to_follower(self, from_, to):
        log.info("peer [%s] entering FOLLOWER state", self.candidate_id)
        self._update_memo("vote_due", 0)
        self._update_memo("ping_due", 0)
        self._reset_election_timer()

    @state.transition(FOLLOWER, CANDIDATE)
    def _follower_to_candidate(self, from_, to):
        self._speed_up()
        self.current_term += 1
        self.voted_for = self.candidate_id
        self._update_memo("voted_for_us", False)
        self._update_memo("ping_due", 0)
        self._update_memo("vote_due", 0)
        self._reset_election_timer()
        log.info(
            "peer [%s] moving from FOLLOWER -> CANDIDATE, "
            "starting new election for term %s, timeout %.4f",
            self.candidate_id,
            self.current_term, self.election_timer.interval)

    @state.transition(CANDIDATE, CANDIDATE, interesting=False)
    def _candidate_to_candidate(self, from_, to):
        self.current_term += 1
        self._update_memo("voted_for_us", False)
        self._update_memo("vote_due", 0)
        self._reset_election_timer()
        log.debug(
            "peer [%s] remains in CANDIDATE, starting election "
            "for new term %s, timeout %.4f",
            self.candidate_id, self.current_term, self.election_timer.interval)

    @state.transition(CANDIDATE, LEADER)
    def _candidate_to_leader(self, from_, to):
        self.current_leader = self.candidate_id
        log.info(
            "peer [%s] moving from CANDIDATE -> LEADER", self.candidate_id)
        self._update_memo("ping_due", 0)

    @state.transition(CANDIDATE, FOLLOWER)
    def _candidate_to_follower(self, from_, to, term):
        log.info(
            "peer [%s] moving from CANDIDATE -> FOLLOWER", self.candidate_id)
        self.voted_for = None
        self.current_term = term

    @state.transition(LEADER, FOLLOWER)
    def _leader_to_follower(self, from_, to, term):
        log.info(
            "peer [%s] moving from LEADER -> FOLLOWER", self.candidate_id)
        self.voted_for = None
        self.current_term = term

    state.null(FOLLOWER, FOLLOWER)

    # the construction of these methods is taken more or less directly
    # from https://github.com/ongardie/raftscope/blob/master/raft.js
    def request_append_entries(self, peer, now):
        log_rec = AppendEntries(
            self.current_term,
            self.candidate_id,
            0, 0, [], 0  # log values, fixed for an empty log
        )

        peer.memos['ping_due'] = now + self.rpc_timeout

        log_rec.send_async(
            peer.rpc_service,
            lambda response: self.handle_append_entries_response(
                log_rec, peer, response),
            lambda response: self.error_response(log_rec, peer, response),
            timeout=self.rpc_timeout
        )

    def handle_append_entries_request(self, request, peer):
        """Receive an AppendEntries RPC call."""

        success = False

        # they're ahead of us, we step down
        if self.current_term < request.term and self.LEADER.current(self):
            self.FOLLOWER.go(self, request.term)

        if self.current_term == request.term:
            if self.CANDIDATE.current(self):
                self.FOLLOWER.go(self, request.term)
            self.election_timer.reset(time.time())
            success = True
            self.current_leader = request.leader_id

        return self.current_term, success

    def handle_append_entries_response(self, append_entries, peer, response):
        peer.memos['ping_due'] = \
            time.time() + self.election_timeout_range[-1] / 2

        response_term, response_success = response

        if self.current_term < response_term:
            self.FOLLOWER.go(self, response_term)

    def request_vote_request(self, peer, now):
        # vote for self
        assert self.current_term
        vote = RequestVote(
            self.current_term,
            self.candidate_id,
            0, 0
        )
        peer.memos['vote_due'] = now + self.rpc_timeout
        vote.send_async(
            peer.rpc_service,
            lambda response: self.handle_request_vote_response(
                vote, peer, response),
            lambda response: self.error_response(vote, peer, response),
            timeout=self.rpc_timeout,
        )

    def handle_request_vote_request(self, request, peer):
        granted = False

        if self.current_term < request.term:
            self.FOLLOWER.go(self, request.term)

        if self.current_term == request.term and (
            self.voted_for is None or
            self.voted_for == request.candidate_id
        ):
            granted = True
            self.voted_for = peer.name

        return self.current_term, granted

    def handle_request_vote_response(self, vote_request, peer, response):
        peer.memos['vote_due'] = float("inf")

        response_term, response_granted = response
        if self.current_term < response_term:
            self.FOLLOWER.go(self, response_term)

        if self.CANDIDATE.current(self) and self.current_term == response_term:
            if response_granted:
                peer.memos['voted_for_us'] = True
                self._check_votes()

    def _check_votes(self):
        vote_count = self._get_vote_count()
        if vote_count + 1 > self.p2p.total_number_of_peers / 2:
            self.LEADER.go(self)

    def _get_vote_count(self):
        return sum([
            1 if peer.memos['voted_for_us'] else 0
            for peer in self.p2p.peers
        ])

    def error_response(self, rpc_msg, peer, exception):
        log.info(
            "Exception received from peer %s: %s",
            peer.their_nodename, exception)


class RaftEvents(util.EventListener):
    _dispatch_target = RaftHandler

    def state_changed(self, raft, old_status, new_status):
        pass


class RaftConsoleCommands(object):
    name = 'raft'

    def __init__(self, console):
        console.register_cmds_from(self)
        p2p.P2PConsoleCommands(console)
        self.rpc_service = console.speak_rpc(rpc_reg)

    def cmd_leader(self):
        return CurrentLeader().send(self.rpc_service)

rpc_reg = rpc.RPCReg()


@rpc_reg.call()
class CurrentLeader(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        raft = service_connection.server.memos['raft']
        return raft.current_leader


# p2p RPC commands, these are bi-directional

@rpc_reg.call('term', 'leader_id', 'prev_log_index',
              'prev_log_term', 'entries', 'leader_commit')
class AppendEntries(p2p.P2PRpc):
    def receive_peer_request(self, rpc, peer):
        raft = peer.handler.service.memos['raft']
        return raft.handle_append_entries_request(self, peer)


@rpc_reg.call('term', 'candidate_id', 'last_log_index', 'last_log_term')
class RequestVote(p2p.P2PRpc):
    def receive_peer_request(self, rpc, peer):
        raft = peer.handler.service.memos['raft']
        return raft.handle_request_vote_request(self, peer)




"""Implement a simple RPC model on top of p2p.
"""
import collections
import re
import logging
import itertools

from . import util
from . import exc

log = logging.getLogger(__name__)


class RPCService(object):

    def __init__(self, rpc_reg, transport, rpc_call_timeout=5):
        self.rpc_reg = rpc_reg
        self.transport = transport
        self._message_id = itertools.count(1)
        self.async_suite = transport.async_suite
        self.rpc_call_timeout = rpc_call_timeout
        self._callbacks = {}
        self._listeners = collections.defaultdict(list)

    def speak_rpc(self, rpc_reg):
        self.rpc_reg = self.rpc_reg.join(rpc_reg)

    def listen(self, fn, evt_cls):
        self._listeners[evt_cls].append(fn)

    def message_received(self, message, expected=None):
        match = re.match(
            "^RPC_(REQ|RESP|ERR|EVT)\((\d+), (.+?)\)\: (.*)", message)
        if not match:
            return
        log.debug("received RPC message %s", message)
        message_type = match.group(1)
        message_id = int(match.group(2))
        rpc_name = match.group(3)
        message = match.group(4)

        if message_type == 'REQ':
            self._service_request(
                message_id, rpc_name, message, expected=expected)
        elif expected:
            raise exc.RPCError("Expected only request of type %s" % expected)
        elif message_type == 'RESP':
            self._service_response(message_id, rpc_name, message)
        elif message_type == 'ERR':
            self._service_error(message_id, rpc_name, message)
        elif message_type == 'EVT':
            self._service_event(message_id, rpc_name, message)
        else:
            assert False

    def _service_request(self, message_id, rpc_name, message, expected=None):
        rpc_cls = self.rpc_reg._calls[rpc_name]

        if expected and not issubclass(expected, rpc_cls):
            self._send_error(message_id, rpc_name, "unexpected message")

        try:
            rpc_obj = rpc_cls(*util.json_loads(message))
        except TypeError as te:
            self._send_error(message_id, rpc_name, str(te))
        else:
            try:
                response = rpc_obj.receive_request(self, self.transport)
            except Exception as err:
                log.error("Error from RPC handler", exc_info=True)
                self._send_error(message_id, rpc_name, err)
            else:
                msg = "RPC_RESP(%s, %s): %s" % (
                    message_id, rpc_name, util.json_dumps(response)
                )
                log.debug("send RPC response %s", msg)
                self.transport.send_message(msg)

    def _service_response(self, message_id, rpc_name, message):
        response = util.json_loads(message)
        waiter = self._callbacks.pop(message_id, None)
        if waiter:
            waiter.set(response)

    def _service_error(self, message_id, rpc_name, message):
        exception = exc.RPCError(message)
        waiter = self._callbacks.pop(message_id, None)
        if waiter:
            waiter.set_exception(exception)

    def _service_event(self, message_id, rpc_name, message):
        rpc_cls = self.rpc_reg._calls[rpc_name]
        try:
            rpc_obj = rpc_cls(*util.json_loads(message))
        except TypeError:
            log.error("Couldn't deserialize rpc event: %s", message)
        else:
            rpc_obj.receive_request(self, self.transport)
            for listener in self._listeners[rpc_cls]:
                listener(rpc_obj)
        return message_id, None, None

    def _send_sync(self, cmd, timeout):
        waiter = self._send_async(cmd, timeout=timeout)
        return waiter.get()

    def _send_async(
            self, cmd, callback=None, exception_callback=None, timeout=None):
        if timeout is None:
            timeout = self.rpc_call_timeout

        message_id = next(self._message_id)

        waiter = self.async_suite.spawn_waiter(
            success=callback,
            failure=exception_callback,
            timeout=timeout,
            on_timeout=lambda: self._callbacks.pop(message_id)
        )
        self._callbacks[message_id] = waiter

        waiter.start()

        self.transport.send_message(
            "RPC_REQ(%s, %s): %s" % (
                message_id, cmd.__class__.__name__, util.json_dumps(cmd))
        )

        return waiter

    def _send_error(self, message_id, rpc_name, errmsg):
        msg = "RPC_ERR(%s, %s): %s" % (
            message_id, rpc_name, errmsg
        )
        log.debug("send RPC error %s", msg)
        self.transport.send_message(msg)

    def _send_event(self, cmd):
        rpc_obj = cmd
        rpc_name = rpc_obj.__class__.__name__
        message_id = next(self._message_id)
        msg = "RPC_EVT(%s, %s): %s" % (
            message_id, rpc_name, util.json_dumps(rpc_obj)
        )
        log.debug("send RPC event %s", msg)
        self.transport.send_message(msg)


class RPCReg(object):
    def __init__(self):
        self._calls = {}

    def join(self, other):
        reg = RPCReg()
        reg._calls.update(self._calls)
        reg._calls.update(other._calls)
        return reg

    def call(self, *fields):

        def decorate(cls):
            new_rpc_type = type(
                cls.__name__,
                (
                    cls,
                    collections.namedtuple(cls.__name__, fields),
                    RPC
                ),
                {}
            )
            self._calls[cls.__name__] = new_rpc_type
            return new_rpc_type
        return decorate


class RPC(tuple):

    def send_async(
            self, rpc, callback, exception_callback=None, timeout=None):
        return rpc._send_async(self, callback, exception_callback, timeout)

    def send(self, rpc, timeout=None):
        return rpc._send_sync(self, timeout=timeout)

    def receive_request(self, rpc, service):
        raise NotImplementedError()

    def _serialize(self):
        return util.json_dumps(self)

    def __nonzero__(self):
        return True


class RPCEvent(RPC):
    def send_async(
            self, rpc, callback, exception_callback=None, timeout=None):
        raise NotImplementedError("Call send() for RPCEvent sending")

    def send(self, rpc):
        rpc._send_event(self)

    def receive_request(self, rpc, service):
        pass

rpc_reg = RPCReg()


@rpc_reg.call('target', 'greeting')
class HelloRPC(RPC):
    """Sample RPC message.

    E.g.::

        msg = rpc.HelloRPC("some nodename", "some greeting")
        response = msg.send(some_rpc_module, some_target)
        print "Got repsonse! %s" % response

    """
    def receive_request(self, rpc, service):
        return ("Well %s to you too, %s!" % (self.greeting, self.target))



import threading
import json


def json_dumps(text):
    return json.dumps(text)


def json_loads(text):
    return json.loads(text)


class memoized_property(object):
    """A read-only @property that is only evaluated once."""

    def __init__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__

    def __get__(self, obj, cls):
        if obj is None:
            return self
        obj.__dict__[self.__name__] = result = self.fget(obj)
        return result

    def _reset(self, obj):
        memoized_property.reset(obj, self.__name__)

    @classmethod
    def reset(cls, obj, name):
        obj.__dict__.pop(name, None)


def periodic_timer(interval, start=0):
    """return True only every 'interval' seconds."""

    last_check = [start]

    def reset(now):
        last_check[0] = now

    def check(now):
        if now - last_check[0] > interval:
            last_check[0] = now
            return True
        else:
            return False
    check.last_check = last_check
    check.interval = interval
    check.reset = reset
    return check


class Startable(object):
    _started = False
    _starting = False

    @memoized_property
    def _startup_mutex(self):
        return threading.Lock()

    def start(self):
        if self._started:
            return
        with self._startup_mutex:
            if self._started:
                return

            self._starting = True
            try:
                self.do_start()
            finally:
                self._started = True
                self._starting = False

    def stop(self):
        if not self._started:
            return
        with self._startup_mutex:
            if not self._started:
                return

            try:
                self.do_stop()
            finally:
                self._started = False

    def do_start(self):
        raise NotImplementedError()

    def do_stop(self):
        raise NotImplementedError()






import logging
import operator

log = logging.getLogger(__name__)


class State(object):
    def __init__(self, machine, name):
        self.machine = machine
        self.name = name

    def __str__(self):
        return "[%s]" % self.name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        try:
            return other.name == self.name
        except AttributeError:
            return False

    def current(self, target):
        inst = statemachine(target)
        return inst.state is self

    def go(self, target, *arg, **kw):
        inst = statemachine(target)
        return inst._go_target(target, self, *arg, **kw)

statemachine = operator.attrgetter("statemachine")


class StateMachine(object):
    def __init__(self):
        self._transitions = {}
        self._symbols = {}
        self._max_symbol_size = 0
        self.START = self.new("START")

    _BOUNCED = object()

    def new(self, sname):
        assert sname not in self._symbols, "name %s already exists" % sname
        self._symbols[sname] = state = State(self, sname)
        self._max_symbol_size = max(self._max_symbol_size, len(sname))
        return state

    def _null_transition(self, instance, from_, to, *arg, **kw):
        pass

    def init_instance(self, target):
        inst = StateMachineInstance(self, target)
        target.statemachine = inst
        return inst

    def add_listener(self, target, listener):
        statemachine(target).add_listener(listener)

    def get(self, target):
        inst = statemachine(target)
        return inst.state

    def str(self, target):
        inst = statemachine(target)
        return str(inst)

    def null(self, from_, to, **kw):
        """Mark a transition as it should pass silently."""
        self.transition(from_, to, **kw)(self._null_transition)

    def bounce(self, from_, to, **kw):
        """Mark a transition as that it should "bounce" back, not move."""
        self.transition(from_, to, **kw)(self._BOUNCED)

    def transition(self, from_, to, interesting=True):
        assert (from_, to) not in self._transitions, \
            "transition %s->%s already exists" % (from_, to)

        def go(fn):
            self._transitions[(from_, to)] = (fn, interesting)
            return fn
        return go


class StateMachineInstance(object):
    __slots__ = 'machine', 'state', 'log', '_transition', '_listeners'

    def __init__(self, machine, target):
        self.machine = machine
        self.state = machine.START
        self.log = logging.getLogger(target.__module__)
        self._transition = False
        self._listeners = []

    def add_listener(self, listener):
        self._listeners.append(listener)

    def _go_target(self, target, state, *arg, **kw):
        try:
            (fn, interesting) = self.machine._transitions[(self.state, state)]

            if fn is self.machine._BOUNCED:
                if interesting:
                    log.debug(
                        "%s: %s -> %s -> bounced -> %s",
                        target, self.state, state, self.state)
                return

        except KeyError:
            raise KeyError("Unknown transition: %s->%s" % (self.state, state))
        else:
            assert not self._transition, "reentrant state change call!"

            old_state = self.state
            if interesting:
                log.debug("%s: %s -> %s", target, old_state, state)
            try:
                self._transition = True
                ret = fn(target, old_state, state, *arg, **kw)
            except Exception:
                raise
            else:
                self.state = state
                for listener in self._listeners:
                    listener(target, old_state, state)
                return ret
            finally:
                self._transition = False

    def __str__(self):
        return str(self.state)





import logging
import time

from . import util
from . import rpc
from . import exc


log = logging.getLogger(__name__)


class RemoteServer(util.Startable):
    def __init__(self, service_config, servicename, nodename):
        self.service_config = service_config
        self.servicename = servicename
        self.nodename = nodename
        self.async_suite = util.async_suite(green=True)
        self.use_auth = self.service_config['use_auth']
        self.memos = {}
        self.rpc_reg = rpc_reg

    def speak_rpc(self, rpc_reg):
        self.rpc_reg = self.rpc_reg.join(rpc_reg)

    def serve_forever(self):
        if self.nodename is None:
            if len(self.service_config['nodes']) > 1:
                raise exc.ServiceException(
                    "--node is required; choose one of %s" %
                    ", ".join(node['name'] for node
                              in self.service_config['nodes']))
            else:
                our_node = self.service_config['nodes'][0]
                self.nodename = our_node['name']
        else:
            nodes = dict(
                (node['name'], node) for node in self.service_config['nodes']
            )
            try:
                our_node = nodes[self.nodename]
            except KeyError:
                raise exc.ServiceException(
                    "no such node name %r" % self.nodename)

        bind_address = our_node['bind']
        bind_port = our_node.get('bind_port', None) or our_node['port']

        self.connections = set()

        if self.service_config['ssl_keyfile']:
            ssl_args = {
                'keyfile': self.service_config['ssl_keyfile'],
                'certfile': self.service_config['ssl_certfile'],
                'do_handshake_on_connect': True
            }
        else:
            ssl_args = {}

        self.server = self.async_suite.server(
            (bind_address, bind_port), self._connected, **ssl_args
        )
        log.info(
            "Listening on %saddress %s port %d",
            "SSL " if ssl_args else "", bind_address, bind_port)

        self.dispatch.before_listen(self)

        self.server.serve_forever()

    def _auth_key_secret(self, key, secret):
        auth_key = self.service_config['auth_key']
        auth_secret = self.service_config['auth_secret']

        return key == auth_key and secret == auth_secret

    def _connected(self, socket, address):
        ClientConnection(self, socket, address)


class ClientConnection(object):
    # TODO: would like to use __slots__ here once we can stablize the API

    _message_terminator = "&&OK"

    def __init__(self, server, socket, address):
        self.server = server
        self.socket = socket
        self.address = address
        self.username = None
        server.connections.add(self)
        log.info("%s", self)
        self.async_suite = server.async_suite
        self.rpc_service = rpc.RPCService(server.rpc_reg, self)
        self._authenticated = False
        self.connected = True
        self.memos = {}
        self.last_ping = util.periodic_timer(Ping.ping_timeout)
        self.last_ping.reset(time.time())
        self.greenlet = self.async_suite.spawn(self._handle, socket)
        self.server.dispatch.client_connection_connected(self.server, self)

    def send_message(self, message):
        if not self.connected:
            raise exc.DisconnectedError("not connected")
        self.socket.sendall(message + self._message_terminator)

    def broadcast_rpc_to_clients(self, msg, filter_fn=None):
        """Send an RPC to all external clients."""

        for server_connection in self.service.connections:
            if filter_fn and not filter_fn(server_connection):
                pass
            msg.send_event(self.rpc_service)

    def _handle(self, socket):
        try:
            for message in self.server.async_suite.receive_messages(
                    self.socket, self._message_terminator,
                    self._check_ping, Ping.ping_interval):

                if not self._authenticated:
                    if not self._authenticate(message):
                        break
                else:
                    self.rpc_service.message_received(message)
        except self.async_suite.interrupted:
            self._close()
        except IOError as ioe:
            self._close(unexpected=True, message=str(ioe))
        else:
            self._close()

    def _check_ping(self):
        if self.last_ping(time.time()):
            raise IOError("ping timeout, no ping from client")

    def _receive_ping(self):
        self.last_ping.reset(time.time())
        return "PONG"

    def _authenticate(self, message):
        self.rpc_service.message_received(
            message, expected=Authenticate)
        if self._authenticated:
            log.info("Authenticated connection from %r", self.username)
            self.server.dispatch.client_connection_authenticated(
                self.server, self)
        else:
            log.info("Authentication failed for username %r", self.username)
        return self._authenticated

    def _cmd_authenticate(self, key, secret):
        self.username = key
        auth = not self.server.use_auth or \
            self.server._auth_key_secret(key, secret)
        if auth:
            self._authenticated = True
            return [
                "Service name: %s" % self.server.servicename
            ]
        else:
            raise exc.AuthFailedError()

    def _cmd_servicename(self):
        return self.server.servicename

    def _cmd_quit(self):
        self.greenlet.kill()

    def __str__(self):
        return "ClientConnection(%s:%s)" % self.address

    def _close(self, unexpected=False, message=None):
        self.server.connections.discard(self)
        if not unexpected:
            self.socket.shutdown(self.async_suite.socket.SHUT_RDWR)
        self.socket.close()
        self.connected = False
        self.server.dispatch.client_connection_disconnected(
            self.server, self, unexpected, message)
        log.info("%s closed (%s)", self, message)


class ServerListener(util.EventListener):
    _dispatch_target = RemoteServer

    def before_listen(self, server):
        pass

    def client_connection_connected(self, server, client_connection):
        pass

    def client_connection_authenticated(self, server, client_connection):
        pass

    def client_connection_disconnected(
            self, server, client_connection, unexpected, message):
        pass

rpc_reg = rpc.RPCReg()


@rpc_reg.call('key', 'secret')
class Authenticate(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        return service_connection._cmd_authenticate(self.key, self.secret)


@rpc_reg.call()
class Ping(rpc.RPC):
    ping_timeout = 30
    ping_interval = 15

    def receive_request(self, rpc, service_connection):
        return service_connection._receive_ping()


@rpc_reg.call()
class Servicename(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        return service_connection._cmd_servicename()


@rpc_reg.call()
class Quit(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        return service_connection._cmd_quit()




"""A P2P "Peer" implementation that rides on top of the
RemoteStatusServer / StatusClient system.

It knows only about sending and receiving data on a TCP socket; other
details as far as connecting, receiving connections, and disconnecting
are handled by the RemoteStatusServer / StatusClient.

"""
from .client import GreenLocalClient
from . import util
from . import rpc
from . import exc

import logging

log = logging.getLogger(__name__)


class PeerRunner(object):
    name = 'p2p'

    def __init__(self, service):
        self.service = service
        self.rpc_reg = rpc_reg

        self.nodename = self.service.nodename
        self.peernames = [
            n['name'] for n in self.service.service_config['nodes']]

        service.speak_rpc(rpc_reg)
        service.memos['p2p'] = self
        util.event_listen(service, "before_listen", self.before_listen)
        util.event_listen(
            service,
            "client_connection_disconnected",
            self.client_connection_disconnected)

        self.peers = set()
        self.async_suite = self.service.async_suite

    def speak_rpc(self, rpc_reg):
        self.rpc_reg = self.rpc_reg.join(rpc_reg)
        self.service.speak_rpc(rpc_reg)

    @property
    def total_number_of_peers(self):
        return len(self.peernames)

    def broadcast_rpc_to_peers(self, msg):
        for peer in self.peers:
            if peer.connected:
                msg.send(peer.rpc_service)

    def _remove_server_peer(self, peer):
        self.peers.remove(peer)
        self.dispatch.peer_removed(self)

    def _peer_error_received(self, peer, exception):
        self.dispatch.peer_exception_received(peer, exception)

    def before_listen(self, service):
        self.nodename = self.service.nodename
        assert self.nodename is not None

        for peername in self.peernames:
            if peername == self.nodename:
                continue
            elif peername > self.nodename:
                peer = ClientPeer(
                    self,
                    self.nodename,
                    self.service.service_config,
                    self.service.servicename, peername
                )
                peer._start()

    def client_connection_disconnected(
            self, server, client_connection, unexpected, message):
        if 'p2p_peer' in client_connection.memos:
            server_peer = client_connection.memos['p2p_peer']
            server_peer.disconnect()

    def _cmd_peers(self):
        peer_dict = dict(
            (name, False) for name in self.peernames
        )
        peer_dict.update(
            (peer.name, peer.connected) for peer in self.peers
        )
        return [
            "%s - %s" %
            (
                nodename,
                "this peer" if nodename == self.nodename
                else "connected" if peer_dict[nodename]
                else "disconnected"
            )
            for nodename in self.peernames
        ]


class PeerListener(util.EventListener):
    _dispatch_target = PeerRunner

    def new_peer(self, peer):
        pass

    def peer_connected(self, peer):
        pass

    def peer_disconnected(self, peer):
        pass

    def peer_removed(self, peer):
        pass

    def peer_exception_received(self, peer, exception):
        pass


class Peer(object):
    """generic 'peer' object that represents p2p operations on a socket.

    Base is agnostic of whether the connection is a client or server
    connection.

    """

    def __init__(self, handler, our_nodename, their_nodename):
        self.handler = handler
        self.async_suite = self.handler.async_suite
        self.our_nodename = our_nodename
        self.their_nodename = self.name = their_nodename
        self.memos = {}

    def _enter_pool(self):
        self.handler.peers.add(self)
        self.handler.dispatch.new_peer(self)


class ServerPeer(Peer):
    """peer that's connected via server.ClientConnection.

    """
    _init = False

    def __init__(
            self, handler, our_nodename, client_connection,
            their_nodename):
        super(ServerPeer, self).__init__(handler, our_nodename, their_nodename)
        self.client_connection = client_connection
        self.client_connection.memos['p2p_peer'] = self

        self.rpc_service = self.client_connection.rpc_service
        log.info(
            "[%s] received peer connection from %s", self.our_nodename,
            self.their_nodename)
        self._enter_pool()
        self._init = True
        self.handler.dispatch.peer_connected(self)

    @property
    def connected(self):
        return self._init and self.client_connection.connected

    def disconnect(self):
        log.info(
            "[%s] received disconnect from [%s]",
            self.our_nodename,
            self.their_nodename)
        self.handler.dispatch.peer_disconnected(self)
        self.handler._remove_server_peer(self)


class ClientPeer(Peer):
    """peer that's connected via client.LocalClient.

    """

    def __init__(
            self, handler, our_nodename, service_config,
            servicename, their_nodename):
        super(ClientPeer, self).__init__(handler, our_nodename, their_nodename)
        self.client = GreenLocalClient.from_config(
            service_config, servicename, their_nodename)
        self.rpc_service = self.client.speak_rpc(handler.rpc_reg)

        self.client.memos['p2p_peer'] = self

        util.event_listen(
            self.client, "client_connected", self.client_connected)
        util.event_listen(
            self.client, "client_disconnected", self.client_disconnected)

        self.connected = False
        self._enter_pool()

    def _start(self):
        log.info(
            "[%s] attempting to establish peer connection to [%s]",
            self.our_nodename, self.their_nodename
        )

        def status(success, exception, time):
            if not success:
                log.info(
                    "[%s] continuing to attempt connection to [%s], "
                    "%d seconds so far (can't connect: %s)",
                    self.our_nodename,
                    self.their_nodename,
                    time,
                    exception
                )

        self.client.connect_persistent(status_fn=status, notify_every=30)

    def client_connected(self, client):
        response = SetAsPeer(self.our_nodename).send(self.rpc_service)
        log.info(
            "[%s] established peer connection to [%s] - %s",
            self.our_nodename, self.their_nodename, response)
        self.connected = True
        self.handler.dispatch.peer_connected(self)

    def client_disconnected(self, client):
        self.connected = False
        self.handler.dispatch.peer_disconnected(self)


class P2PConsoleCommands(object):
    name = 'p2p'

    def __init__(self, console):
        self.console = console
        self.console.register_cmds_from(self)
        self.rpc_service = console.speak_rpc(rpc_reg)

    def cmd_set_as_peer(self, name):
        return SetAsPeer(name).send(self.rpc_service)

    def cmd_peers(self):
        return Peers().send(self.rpc_service)

rpc_reg = rpc.RPCReg()


@rpc_reg.call()
class Peers(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        p2p = service_connection.server.memos['p2p']
        if p2p is None:
            return ["%s - not p2p" % (service_connection.server.nodename, )]
        else:
            return p2p._cmd_peers()


@rpc_reg.call('name')
class SetAsPeer(rpc.RPC):
    def receive_request(self, rpc, service_connection):
        p2p = service_connection.server.memos['p2p']
        ServerPeer(p2p, p2p.nodename, service_connection, self.name)


@rpc_reg.call()
class IsClustered(rpc.RPC):

    def receive_request(self, rpc, service_connection):
        server = service_connection.server
        nodes = server.service_config['nodes']
        is_clustered = len(nodes) > 1 and \
            'p2p' in server.memos
        return is_clustered


class P2PRpc(rpc.RPC):
    def receive_request(self, rpc, service_connection_or_client):
        try:
            peer = service_connection_or_client.memos['p2p_peer']
        except KeyError:
            raise exc.RPCError(
                "Received P2P RPC message from client that didn't "
                "send setaspeer: %s, %s" %
                (service_connection_or_client, self)
            )
        else:
            return self.receive_peer_request(rpc, peer)

    def receive_peer_request(self, rpc, peer):
        raise NotImplementedError()





import socket
import random
from . import rpc
import time

from . import util
from . import server as _server
from . import exc
import logging

log = logging.getLogger(__name__)


class LocalClient(object):
    def __init__(self, servicename, key, secret, node_configs):
        self.servicename = servicename
        self.key = key
        self.secret = secret
        self.async_suite = self._make_async_suite()
        self.rpc_service = rpc.RPCService(_server.rpc_reg, self)
        self.node_configs = node_configs
        self.connected = False
        self._reconnect_on_disconnect = False
        self.last_ping = util.periodic_timer(_server.Ping.ping_timeout)
        self.memos = {}

    def speak_rpc(self, rpc_reg):
        self.rpc_service.speak_rpc(rpc_reg)
        return self.rpc_service

    def _make_async_suite(self):
        return util.async_suite()

    @classmethod
    def from_multi_host_port(cls, username, password, hosts):
        node_configs = [
            {"hostname": hostname, "port": port}
            for hostname, port in hosts
        ]
        return cls(
            "%s:%s" % (hostname, port),
            username,
            password, node_configs
        )

    @classmethod
    def from_host_port(cls, username, password, hostname, port):
        node_configs = [
            {"hostname": hostname, "port": port}
        ]
        return cls(
            "%s:%s" % (hostname, port),
            username,
            password, node_configs
        )

    @classmethod
    def from_config(
            cls, service_config, servicename,
            nodename=None, username=None, password=None):

        if nodename is None:
            our_nodes = service_config['nodes']
        else:
            nodes = dict(
                (node['name'], node) for node in service_config['nodes']
            )
            try:
                our_nodes = [nodes[nodename]]
            except KeyError:
                raise exc.ServiceException(
                    "no such node name %r" % nodename)

        # TODO: if node is none, send list of all hosts
        key = username or service_config['auth_key']
        secret = password or service_config['auth_secret']
        return cls(servicename, key, secret, our_nodes)

    def connect(self):
        nodes = list(self.node_configs)
        random.shuffle(nodes)
        for node in nodes:
            host = node['hostname']
            port = node['port']
            try:
                self.socket = self.async_suite.create_connection(
                    (host, port), timeout=2)
            except IOError as ioe:
                continue
            else:
                self.connected = True
                self.host = host
                self.port = port
                self.nodename = node['name']
                break

        else:
            raise exc.DisconnectedError(
                "Can't connect to servicename=%r: %s" %
                (self.servicename, ioe)
            )

        self._handle_thread = self.async_suite.background_thread(self._handle)

        self._authenticate()
        self.last_ping.reset(time.time())
        self.dispatch.client_connected(self)

    def connect_persistent(self, status_fn=None, notify_every=None):
        self._reconnect_on_disconnect = True
        self.async_suite.spawn(
            self._connect_persistent,
            status_fn=status_fn, notify_every=notify_every)

    def _connect_persistent(self, status_fn=None, notify_every=None):
        now = started = time.time()
        while not self.connected:
            try:
                self.connect()
            except exc.DisconnectedError as de:
                attempting_for = time.time() - now
                if attempting_for > notify_every:
                    now = time.time()
                    status_fn(False, de, time.time() - started)
                self.async_suite.sleep(1)
            else:
                if status_fn:
                    status_fn(True, None, time.time() - started)
                log.info(
                    "LocalClient connected to host %s:%s",
                    self.host, self.port)
                break

    def on_disconnect(self, unexpected, message):
        if self._reconnect_on_disconnect:
            log.info(
                "LocalClient lost connection to host %s:%s; "
                "reconnecting to servicename=%r",
                self.host, self.port, self.servicename)

            def status(success, exception, time):
                log.info(
                    "Continuing to attempt connection to servicename=%r, "
                    "%d seconds so far (%s)",
                    self.servicename,
                    time,
                    exception
                )
            self.connect_persistent(status_fn=status, notify_every=30)
        else:
            log.info(
                "LocalClient lost connection to host %s:%s",
                self.host, self.port)

    def _handle(self):
        try:
            for message in self.async_suite.receive_messages(
                    self.socket, self._message_terminator,
                    self._check_ping, _server.Ping.ping_interval):
                self.rpc_service.message_received(message)
        except IOError as ioe:
            if self.connected:
                self._close(unexpected=True, message=str(ioe))

    def _authenticate(self):
        try:
            _server.Authenticate(
                self.key, self.secret).send(self.rpc_service)
        except exc.RPCError:
            raise exc.AuthFailedError()

    _message_terminator = "&&OK"

    def _check_ping(self):
        if self.last_ping(time.time()):
            self._close(
                unexpected=True, message="ping timeout, no pong from server")
        else:
            _server.Ping().send_async(
                self.rpc_service,
                callback=self._receive_ping,
                exception_callback=self._ping_failed)

    def _receive_ping(self, message):
        self.last_ping.reset(time.time())

    def _ping_failed(self, exception):
        log.error("ping failed: %s", exception)

    def close(self):
        if self.connected:
            self.connected = False
            self.socket.shutdown(socket.SHUT_RDWR)
            self.socket.close()

    def _close(self, unexpected=False, message=None):
        was_connected = self.connected
        self.connected = False
        if not unexpected:
            self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()
        if was_connected:
            # note: the console looks for this
            # in its on_disconnect handler, so set first
            self.dispatch.client_disconnected(self)
            self.on_disconnect(unexpected, message)

    def send_message(self, message):
        if not self.connected:
            raise exc.DisconnectedError("client is not connected")
        # print "send message", message
        self.socket.sendall(message + self._message_terminator)


class GreenLocalClient(LocalClient):
    def _make_async_suite(self):
        return util.async_suite(green=True)


class ClientListener(util.EventListener):
    _dispatch_target = LocalClient

    def client_connected(self, client):
        pass

    def client_disconnected(self, client):
        pass





from logging.config import fileConfig
import re

from .util import compat
from . import exc


class Config(object):
    def __init__(self, prefix="zk_", config_elements=None):
        self.services = {}
        self.sources = []
        self.prefix = prefix
        self.config_elements = list(server_config_names)
        if config_elements:
            self.config_elements.extend(config_elements)

    def add_config(self, config):
        for section in config.sections():
            if section.startswith("%sservice" % self.prefix):
                self.add_service_config(dict(config.items(section)))

    def add_config_from_config_string(self, string):
        file_config = compat.SafeConfigParser()
        file_config.readfp(compat.StringIO(string))
        self.sources.append("<string>")
        self.add_config(file_config)

    def add_config_from_config_file(self, fname):
        file_config = compat.SafeConfigParser()
        file_config.read(fname)
        self.add_config(file_config)
        self.sources.append(fname)

    def load_logging_configs(self):
        for fname in self.sources:
            self._attempt_logging_config(fname)

    def _attempt_logging_config(self, fname):
        try:
            fileConfig(fname, disable_existing_loggers=False)
        except compat.configparser.NoSectionError:
            pass

    def add_service_config(self, service_cfg):
        cfg_dict = {}
        for element in self.config_elements:
            if element.name in service_cfg:
                raw_value = service_cfg[element.name]
                element.apply_fixer(raw_value, cfg_dict)
            else:
                element.apply_default(cfg_dict)
        self.services[cfg_dict['name']] = cfg_dict

    @classmethod
    def from_config_string(cls, string, prefix="zk_", config_elements=None):
        config = Config(prefix=prefix, config_elements=config_elements)
        config.add_config_from_config_string(string)
        return config

    @classmethod
    def from_config_file(cls, fname, prefix="zk_", config_elements=None):
        config = Config(prefix=prefix, config_elements=config_elements)
        config.add_config_from_config_file(fname)
        return config

    def config_for_servicename(self, service_name):
        try:
            service_cfg = self.services[service_name]
        except KeyError:
            raise exc.ConfigurationException(
                "No such service config: %r" % service_name)
        else:
            return service_cfg


class Fixer(object):

    def apply_to_dict(self, key, fixed_value, dest_dict):
        dest_dict[key] = fixed_value

    def fix_value(self, value):
        if value == 'none':
            value = None

        if value is not None:
            value = self.fix_non_none_value(value)
        return value

    def fix_non_none_value(self, value):
        return value

    def __call__(self, key, raw_value, dest_dict):
        self.apply_to_dict(
            key, self.fix_value(raw_value), dest_dict
        )


class BoolFixer(Fixer):
    def fix_non_none_value(self, value):
        if value.lower() in ('true', 1, 'yes'):
            return True
        elif value.lower() in ('false', 0, 'no'):
            return False
        else:
            raise ValueError("Invalid boolean string: %r" % value)


class IntFixer(Fixer):
    def fix_non_none_value(self, value):
        return int(value)


class StringListFixer(Fixer):
    def fix_non_none_value(self, value):
        return list(
            [e for e in
                [elem.strip() for elem in re.split(r', +', value)]
                if e])


class HostFixer(Fixer):
    def __init__(self, default_port=None):
        self.default_port = default_port

    def fix_value(self, value):
        lines = re.split("\n", value.strip())
        hosts = []
        for line in lines:
            # skip commented lines
            if re.match('^\s*#', line):
                continue

            host_rec = {}
            tokens = re.split("(?:\s|(=))+", line.strip())

            host_rec['name'] = tokens.pop(0)

            while tokens:
                token_key = tokens.pop(0)
                if tokens and tokens[0] == "=":
                    tokens.pop(0)
                    try:
                        token_value = tokens.pop(0)
                    except IndexError:
                        raise exc.ConfigurationException(
                            "Expected value after symbol %r: %r" %
                            (token_key, line))
                    else:
                        if token_key == 'hostname':
                            port_key = 'port'
                        elif token_key == 'bind':
                            port_key = 'bind_port'
                        else:
                            port_key = None

                        if port_key:
                            if ':' in token_value:
                                host, port = token_value.split(":", 1)
                                port = IntFixer().fix_value(port)
                                host_rec[port_key] = port
                            elif port_key == 'port':
                                host = token_value
                                host_rec[port_key] = self.default_port
                            elif port_key == 'bind_port':
                                host = token_value
                            host_rec[token_key] = host

                elif token_key is not None:
                    host_rec[token_key] = True

            if 'hostname' not in host_rec:
                raise exc.ConfigurationException(
                    "hostname is required in host record: %r" % line)

            if 'bind' not in host_rec:
                host_rec['bind'] = '0.0.0.0'
            if 'bind_port' not in host_rec:
                host_rec['bind_port'] = host_rec['port']
            hosts.append(host_rec)
        return hosts


class ConfigElement(object):
    __slots__ = 'name', 'fixer', 'default'

    def __init__(self, name, fixer=Fixer, default=None, fixer_defaults={}):
        self.name = name
        self.fixer = fixer(**fixer_defaults)
        self.default = default

    def apply_fixer(self, raw_value, cfg_dict):
        self.fixer(self.name, raw_value, cfg_dict)

    def apply_default(self, cfg_dict):
        if self.default is REQUIRED:
            raise exc.ConfigurationException(
                "Argument %r is required" % self.name)
        self.fixer.apply_to_dict(self.name, self.default, cfg_dict)

REQUIRED = object()

server_config_names = [
    ConfigElement('name', default=REQUIRED),
    ConfigElement(
        'nodes', HostFixer, default=REQUIRED,
        fixer_defaults={"default_port": 5084}),
    ConfigElement('use_auth', BoolFixer, False),
    ConfigElement('auth_key', default=None),
    ConfigElement('auth_secret', default=None),
    ConfigElement('cluster_servers', BoolFixer, True),
    ConfigElement('ssl_keyfile'),
    ConfigElement('ssl_certfile'),
]


_CFG = Config()




import re
import cmd
import sys

from .util import compat
from . import exc
from . import client as _client


class ConsoleCmds(cmd.Cmd):
    def __init__(self, console):
        cmd.Cmd.__init__(self)
        self.console = console

    def _tokenize(self, cmd):
        return [
            token for token in
            re.split(r'\s+', cmd)
            if token
        ]

    def __getattr__(self, key):
        if key.startswith("do_"):
            cmd_name = key[3:]
            try:
                cmd = self.console.cmd.commands[cmd_name]
            except KeyError:
                raise AttributeError(key)

            def go(cmd_):
                self.console._invoke_cmd(cmd)
            return go
        else:
            raise AttributeError(key)

    def do_help(self, cmd):
        self.console._handle_response(
            self.console.cmd.help()
        )

    def emptyline(self):
        pass

    def do_EOF(self, cmd):
        self.do_quit(cmd)

    def postcmd(self, stop, line):
        return not self.console.connected


class Console(_client.LocalClient):
    def __init__(self, *arg, **kw):
        super(Console, self).__init__(*arg, **kw)
        self.console = ConsoleCmds(self)
        self.cmd = _CmdInterface()
        self.cmd.commands_from(self)

    def register_cmds_from(self, source):
        self.cmd.commands_from(source)

    def cmd_servicename(self):
        """Return the service name."""
        from . import server
        return server.Servicename().send(self.rpc_service)

    def cmd_quit(self):
        """Quit the console."""

        self.close()

    def on_disconnect(self, unexpected=False, message=None):
        self.output("Connection closed (%s)" % message)
        self.console.connected = False
        sys.exit(-1 if unexpected else 0)

    def _invoke_cmd(self, cmd, args=(), as_grid=False, as_list=False):
        try:
            response = cmd(*args)
        except exc.RPCError as err:
            self.error(err)
        else:
            return self._handle_response(
                response, as_grid=as_grid, as_list=as_list)

    def _handle_response(self, response, as_grid=False, as_list=False):
        err = False

        if response:
            if not as_grid and isinstance(response, list):
                as_list = True

            if not as_grid and not as_list:
                self.output(compat.text_type(response))
                return True

            if as_grid:
                grid = _render_as_grid()

            while response:
                item = response.pop(0)
                if as_grid:
                    grid.append(item)
                else:
                    self.output(compat.text_type(item))

            if as_grid:
                self.output(compat.text_type(grid))

        return not err

    def interact(self):
        self.console.cmdloop()

    def output(self, text, *args):
        self.console.stdout.write(text % args + "\n")

    def error(self, message):
        self.output("Error: %s", message)

    def cmd_help(self):
        """Command help"""
        self.output(
            "\n".join([
                "Command help:",
            ] + self.cmd._help)
        )


class _CmdInterface(object):
    def __init__(self):
        self._help = []
        self.commands = {}

    def commands_from(self, obj):
        # TOOD: make this interact with the console, with the help,
        # etc.
        for key in dir(obj):
            if key.startswith("cmd_"):
                self.add_command(getattr(obj, key))

    def add_command(self, cmd):
        name = cmd.__name__
        assert name.startswith("cmd_")
        self.commands[name[4:]] = cmd
        self._help.insert(0, "%s - %s" % (name[4:], cmd.__doc__))

    def __getattr__(self, name):
        try:
            return self.commands[name]
        except KeyError:
            raise exc.CommandError("no such command '%s'" % name)


class _render_as_grid(object):
    def __init__(self):
        self.grids = []
        self.current_grid = None

    def append(self, record):
        if not self.current_grid:
            self._create_grid(record)
        elif record.keys() != self.current_grid['keys']:
            self._create_grid(record)

        row = tuple(str(record[key]) for key in self.current_grid['keys'])
        self.current_grid['records'].append(row)
        self.current_grid["max_lengths"] = [
            max(current_length, len(item))
            for item, current_length
            in zip(row, self.current_grid["max_lengths"])
        ]

    def _create_grid(self, record):
        self.current_grid = grid = {
            "records": [],
            "keys": record.keys(),
            "max_lengths": [len(key) for key in record.keys()]
        }
        self.grids.append(grid)

    def __str__(self):
        return "\n".join(
            self._render_grid(grid) for grid in self.grids
        )

    def _render_grid(self, grid):
        max_lengths = grid["max_lengths"]

        tmpl = "    ".join(
            "%%%d.%ds" % (len_, len_) for len_ in max_lengths)

        return "\n".join(
            [tmpl % tuple(grid["keys"])] +
            [
                (tmpl % row)
                for row in grid["records"]
            ]
        )




from .async import async_suite  # noqa
from .async import stop_on_keyinterrupt  # noqa
from .event import EventListener  # noqa
from .event import event_listen  # noqa
from .util import Startable  # noqa
from .util import memoized_property  # noqa
from .util import periodic_timer  # noqa
from .util import json_dumps  # noqa
from .util import json_loads  # noqa




from .util import memoized_property
from .util import periodic_timer

import contextlib
import time
import threading
import sys


@contextlib.contextmanager
def stop_on_keyinterrupt():
    try:
        yield
    except KeyboardInterrupt:
        sys.exit()


class AsyncWaiter(object):
    __slots__ = (
        'factory', 'success', 'failure', 'timeout',
        'on_timeout', 'result', 'exception', 'event',
        'thread'
    )

    NO_RESULT = object()

    def __init__(self, factory, timeout=None, on_timeout=None):
        self.factory = factory
        self.timeout = timeout
        self.on_timeout = on_timeout
        self.event = factory.event()
        self.thread = None
        self.result = self.NO_RESULT
        self.exception = None

    def get(self):
        if self.thread is not None:
            self.thread.join()
        if self.exception:
            raise self.exception
        elif self.result is self.NO_RESULT:
            raise Exception(
                "No result loaded within timeout %s" % self.timeout)
        else:
            return self.result

    def set(self, value):
        self.result = value
        self.event.set()

    def set_exception(self, exception):
        self.exception = exception
        self.event.set()

    def start(self):
        self.thread = self.factory.spawn(self._wait)

    def _wait(self):
        if not self.event.wait(self.timeout):
            self.on_timeout()


class CallbackAsyncWaiter(AsyncWaiter):
    __slots__ = ('success', 'failure')

    def __init__(
            self, factory, success=None, failure=None, timeout=None,
            on_timeout=None):
        super(CallbackAsyncWaiter, self).__init__(factory, timeout, on_timeout)
        self.success = success
        self.failure = failure

    def get(self):
        raise NotImplementedError("can't call get() on a callback waiter")

    def _wait(self):
        if not self.event.wait(self.timeout):
            self.on_timeout()
        elif self.success and self.result is not self.NO_RESULT:
            self.success(self.result)
        elif self.failure and self.exception:
            self.failure(self.exception)


class AsyncSuite(object):
    @classmethod
    def get(cls, green=False):
        """Return a "suite" of async-sensitive utilities.

        This is a one-stop place to get a simple namespace with all those
        system calls that change w/ eventlet context.

        """
        if green:
            return _GreenAsyncSuite()
        else:
            return _ThreadedAsyncSuite()

    def create_connection(self, addr, timeout=None):
        return self.socket.create_connection(addr, timeout=timeout)

    def select_socket(self, socket, timeout):
        return self.select.select([socket], [], [], timeout)

    def background_thread(self, fn):
        def go():
            with stop_on_keyinterrupt():
                fn()

        thread = self.thread(go)
        thread.daemon = True
        thread.start()
        return thread

    def receive_messages(
            self, socket, message_terminator, ping_handler, ping_interval):

        ping_timer = periodic_timer(ping_interval)

        mtl = len(message_terminator)

        buf = []
        while True:

            try:
                ready_to_read, ready_to_write, in_error = \
                    self.select_socket(socket, int(ping_interval / 4))
            except self.select_error as se:
                raise IOError("Select error: %s" % se)

            if ready_to_read:
                data = socket.recv(1024)
                if data == '':
                    raise IOError("no more data")
                while True:
                    mti = data.find(message_terminator)
                    if mti != -1:
                        buf.append(data[0:mti])
                        data = data[mti + mtl:]
                        # print "receive message", ''.join(buf)
                        yield ''.join(buf)
                        buf[:] = []
                    else:
                        buf.append(data)
                        break

                continue

            if ping_timer(time.time()):
                # `print "running ping handler"
                ping_handler()

            # allow pre-emption in case select() is broken
            self.sleep(.01)

    def spawn_waiter(
            self, success=None, failure=None, timeout=None, on_timeout=None):
        if success:
            return CallbackAsyncWaiter(
                self, success, failure, timeout, on_timeout)
        else:
            return AsyncWaiter(self, timeout, on_timeout)

    def event(self):
        raise NotImplementedError()

    def sleep(self):
        raise NotImplementedError()

    def spawn(self, fn):
        raise NotImplementedError()

    def queue(self):
        raise NotImplementedError()

    @memoized_property
    def socket(self):
        raise NotImplementedError()

    @memoized_property
    def select(self):
        raise NotImplementedError()

    @memoized_property
    def select_error(self):
        return self.select.error

async_suite = AsyncSuite.get


class _ThreadedAsyncSuite(AsyncSuite):
    def event(self):
        return threading.Event()

    def sleep(self, duration):
        return time.sleep(duration)

    def spawn(self, fn, *args, **kw):
        t = threading.Thread(target=fn, args=args, kwargs=kw)
        t.start()
        return t

    def thread(self, fn):
        return threading.Thread(target=fn)

    def lock(self):
        return threading.Lock()

    @memoized_property
    def queue(self):
        import Queue
        return Queue.Queue

    @memoized_property
    def socket(self):
        import socket
        return socket

    @memoized_property
    def select(self):
        import select
        return select

    def server(self, address, callback, **ssl_args):
        raise NotImplementedError()


class _GreenAsyncSuite(AsyncSuite):
    def __init__(self):
        import eventlet as _et
        import eventlet.event as _et_event
        import eventlet.green as _et_green
        self._eventlet = _et
        self._eventlet_green = _et_green
        self._eventlet_event = _et_event

    def spawn(self, fn, *args, **kw):
        # we need a traditional threading API, e.g.
        # join()
        thread = self.threading.Thread(target=fn, args=args, kwargs=kw)
        thread.start()
        return thread

    @memoized_property
    def threading(self):
        from eventlet.green import threading
        return threading

    @memoized_property
    def queue(self):
        from eventlet import queue
        return queue.Queue

    def thread(self, fn):
        return self.threading.Thread(target=fn)

    @memoized_property
    def interrupted(self):
        import greenlet
        return greenlet.GreenletExit

    def event(self):
        from eventlet.green import threading
        return threading.Event()

    def sleep(self, duration):
        return self._eventlet.sleep(duration)

    def lock(self):
        return self.threading.Lock()

    @memoized_property
    def socket(self):
        return self._eventlet_green.socket

    @memoized_property
    def select(self):
        return self._eventlet_green.select

    def server(self, address, callback, **ssl_args):
        return _EventletServer(self, address, callback, **ssl_args)


class _EventletServer(object):
    def __init__(self, async_suite, address, callback, **ssl_args):
        if ssl_args:
            raise NotImplementedError("Ssl not supported yet")
        self.async_suite = async_suite
        self.address = address
        self.callback = callback

    def serve_forever(self):
        server = self.async_suite._eventlet.listen(self.address)
        with stop_on_keyinterrupt():
            while True:
                try:
                    socket, address = server.accept()
                except OSError:
                    pass
                else:
                    self.async_suite.spawn(self.callback, socket, address)




class ServiceException(Exception):
    pass


class CommandError(ServiceException):
    pass


class DisconnectedError(ServiceException):
    pass


class AuthFailedError(ServiceException):
    pass


class RPCError(ServiceException):
    pass


class ConfigurationException(ServiceException):
    pass


import sys
py3k = sys.version_info >= (3, 0)

if py3k:
    string_types = str,
    binary_type = bytes
    text_type = str
else:
    string_types = basestring,
    binary_type = str
    text_type = unicode


if py3k:
    from configparser import ConfigParser as SafeConfigParser
    import configparser
else:
    from ConfigParser import SafeConfigParser
    import ConfigParser as configparser


if py3k:
    from io import StringIO
else:
    # accepts strings
    from StringIO import StringIO

if py3k:
    import queue
else:
    import Queue as queue

if py3k:
    raw_input = input
else:
    raw_input = raw_input


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass.

    Drops the middle class upon creation.

    Source: http://lucumr.pocoo.org/2013/5/21/porting-to-python-3-redux/

    """

    class metaclass(meta):
        __call__ = type.__call__
        __init__ = type.__init__

        def __new__(cls, name, this_bases, d):
            if this_bases is None:
                return type.__new__(cls, name, (), d)
            return meta(name, bases, d)
    return metaclass('temporary_class', None, {})



from setuptools import setup
from setuptools import find_packages
import os
import re


with open(
        os.path.join(
            os.path.dirname(__file__),
            'zkcluster', '__init__.py')) as v_file:
    VERSION = re.compile(
        r".*__version__ = '(.*?)'",
        re.S).match(v_file.read()).group(1)

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as r_file:
    readme = r_file.read()


requires = [
    'eventlet'
]

setup(
    name='zkcluster',
    version=VERSION,
    description="zzzeek's clustering server framework",
    long_description=readme,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
    ],
    author='Mike Bayer',
    author_email='mike@zzzcomputing.com',
    url='http://bitbucket.org/zzzeek/zkcluster',
    license='MIT',
    packages=find_packages('.', exclude=['test*']),
    include_package_data=True,
    tests_require=['mock'],
    zip_safe=False,
    install_requires=requires,
    entry_points={
    }
)



import argparse
import getpass

from . import config as _config
from . import exc
from . import console as _console
from . import server as _server


class Cmd(object):

    def create_subparser(self, parser, subparsers):
        raise NotImplementedError()

    def setup_arguments(self, parser):
        raise NotImplementedError()

    def go(self, cmdline):
        raise NotImplementedError()


class ListenCmd(Cmd):
    def create_subparser(self, parser, subparsers):
        return subparsers.add_parser(
            "listen",
            help="listen for connections")

    def setup_arguments(self, parser):
        parser.add_argument(
            "-s", "--service", type=str,
            default="default", help="name of service")
        parser.add_argument(
            "-n", "--node", type=str, help="name of service node")
        return parser

    def go(self, cmdline):
        servicename = cmdline.options.service
        nodename = cmdline.options.node

        server_config = cmdline.config.config_for_servicename(servicename)
        server = self.get_server(server_config, servicename, nodename)
        self.init_server(server)
        server.serve_forever()

    def get_server(self, server_config, servicename, nodename):
        return _server.RemoteServer(server_config, servicename, nodename)

    def init_server(self, server):
        pass


class ClientCmd(Cmd):
    def setup_arguments(self, parser):
        parser.add_argument(
            "-s", "--service", type=str,
            default="default",
            help="name of service to connect (exclusive against --host)")
        parser.add_argument(
            "-H", "--host", type=str,
            help="host[:port] to connect (exclusive against --service)"
        )
        parser.add_argument(
            "-n", "--node", type=str,
            help="specific node name to connect to (assumes --service)")
        parser.add_argument(
            "-u", "--user", type=str, help="username")
        parser.add_argument(
            "-p", "--password", type=str, help="password")

    def get_client_args(self, options, config):

        if options.host:
            if options.service not in (None, 'default'):
                raise exc.ServiceException(
                    "Can't send --service argument with --host")
            if options.node:
                raise exc.ServiceException(
                    "Can't send --node argument with --host"
                )
            if ":" in options.host:
                hostname, port = options.host.split(":", 1)
                port = int(port)
            else:
                hostname = options.host
                port = 5800

            if options.user and not options.password:
                password = getpass.getpass("Password:")
            else:
                password = options.password

            return {
                "user": options.user, "password": password,
                "hostname": hostname, "port": port}

        elif options.service:
            service_config = config.config_for_servicename(options.service)
            if options.user and not options.password:
                password = getpass.getpass("Password:")
            else:
                password = options.password

            return {
                "service_config": service_config,
                "service": options.service,
                "node": options.node, "user": options.user,
                "password": password
            }


class ConsoleCmd(ClientCmd):

    def create_subparser(self, parser, subparsers):
        return subparsers.add_parser(
            "console",
            help="console to a service")

    def go(self, cmdline):

        args = self.get_client_args(cmdline.options, cmdline.config)

        if "host" in args:

            client = self.get_console_host_port(
                args['user'],
                args['password'],
                args['hostname'],
                args['port'])

        elif "service" in args:

            client = self.get_console_servicename(
                args['service_config'], args['service'],
                args['node'], args['user'], args['password'])

        self.init_console(client)

        try:
            client.connect()
        except exc.AuthFailedError:
            print("auth failed")
        except exc.DisconnectedError as de:
            print(de)
        else:
            self.console_connected(client)
            self.run_console(client)

    def run_console(self, client):
        client.interact()

    def get_console_servicename(
            self, server_config, servicename, nodename,
            username=None, password=None):
        return _console.Console.from_config(
            server_config, servicename, nodename,
            username=username, password=password)

    def get_console_host_port(self, username, password, hostname, port):
        return _console.Console.from_host_port(
            username, password, hostname, port)

    def console_connected(self, console):
        pass


class CmdLineBase(object):
    def create_parser(self):
        self.parser = argparse.ArgumentParser(prog=self.prog)
        self.parser.add_argument(
            "-c", "--config",
            type=str,
            help="Path to config file")

    def load_config(self):
        if self.options.config:
            self.config = _config.Config.from_config_file(self.options.config)
        else:
            self.config = _config._CFG

        self.config.load_logging_configs()

    def run(self):
        raise NotImplementedError()

    def main(self, argv=None, prog=None):
        self.options = self.parser.parse_args(argv)

        self.load_config()

        self.run(self.options, self.config)


class SingleCmdLine(CmdLineBase):
    def __init__(self, cmd, prog=None):
        self.prog = prog
        self.cmd = cmd
        self.create_parser()
        cmd.setup_arguments(self.parser)

    def run(self, options, config):
        self.cmd.go(self)


class CmdLine(CmdLineBase):
    def __init__(self, cmds=[ListenCmd(), ConsoleCmd()], prog=None):
        self.prog = prog
        self.create_parser()
        self.subparsers = self.parser.add_subparsers()

        for cmd in cmds:
            self.add_cmd(cmd)

    def add_cmd(self, cmd):
        subparser = cmd.create_subparser(self.parser, self.subparsers)
        cmd.setup_arguments(subparser)
        subparser.set_defaults(cmd=cmd)

    def run(self, options, config):
        cmd = options.cmd
        cmd.go(self)

if __name__ == '__main__':
    CmdLine().main()



from . import compat


def event_listen(target, event_name, fn):
    getattr(
        target.dispatch,
        event_name
    ).listeners.append(fn)


def _is_event_name(name):
    return not name.startswith('_') and name != 'dispatch'


class _EventMeta(type):
    def __init__(cls, classname, bases, dict_):
        _create_dispatcher(cls, classname, bases, dict_)
        return type.__init__(cls, classname, bases, dict_)


class _Dispatch(object):
    pass


class _DispatchAttr(object):
    __slots__ = 'key',

    def __init__(self, key):
        self.key = key

    def __get__(self, obj, cls):
        ret = _DispatchAttrInst()
        obj.__dict__[self.key] = ret
        return ret


class _DispatchAttrInst(object):
    __slots__ = 'listeners',

    def __init__(self):
        self.listeners = []

    def __call__(self, *arg, **kw):
        for fn in self.listeners:
            fn(*arg, **kw)


def _create_dispatcher(cls, classname, bases, dict_):
    """Create a :class:`._Dispatch` class corresponding to an
    :class:`.Events` class."""

    dispatch_base = _Dispatch

    event_names = [k for k in dict_ if _is_event_name(k)]
    dispatch_cls = type("%sDispatch" % classname,
                        (dispatch_base, ), {})

    dispatch_cls._event_names = event_names

    for k in dispatch_cls._event_names:
        setattr(dispatch_cls, k, _DispatchAttr(k))

    if getattr(cls, '_dispatch_target', None):
        cls._dispatch_target.dispatch = _dispatcher(dispatch_cls)


class _dispatcher(object):
    def __init__(self, event_cls):
        self.event_cls = event_cls

    def __get__(self, obj, cls):
        if obj is None:
            return self.dispatch_cls
        obj.__dict__['dispatch'] = disp = self.event_cls()
        return disp


class EventListener(compat.with_metaclass(_EventMeta, object)):
    """Base class for propagated events.

    Basically is the same as the SQLAlchemy event base; this is a super-minimal
    very basic version of the same thing to avoid SQLAlchemy as a dependency.


    """
    # _dispatch_target = <class>


